import pytorch_lightning as pl
import torch
from tango.common import Registrable


class DataModule(pl.LightningDataModule, Registrable):
    def train_dataloader(self) -> torch.utils.data.DataLoader:
        raise NotImplementedError

    def val_dataloader(self) -> torch.utils.data.DataLoader:
        raise NotImplementedError

    def test_dataloader(self) -> torch.utils.data.DataLoader:
        raise NotImplementedError



import json
import pickle
from pathlib import Path
from typing import Any, Dict, List

import chardet
import jsonlines
import pandas as pd


def safe_pd_read_csv(path: str | Path, **kwargs: Dict[str, Any]) -> pd.DataFrame:
    """
    Reads a CSV file with `pd.read_csv` and automatically detects the encoding of the file.
    Parameters:
        path: Path to the CSV file to read.
        encoding: Encoding of the CSV file. If `None`, the encoding will be detected automatically.
        kwargs: Keyword arguments to be passed to `pd.read_csv`.
    Returns:
        A pandas DataFrame.
    """
    try:
        return pd.read_csv(path, **kwargs)
    except UnicodeDecodeError:
        encoding = get_file_encoding(path)
        return pd.read_csv(path, encoding=encoding, **kwargs)


def get_file_encoding(file_path: str | Path) -> str | None:
    with open(file_path, 'rb') as f:
        result = chardet.detect(f.read())
    return result['encoding']


def load_jsonlines(path: str | Path) -> List[Dict]:
    data_list = []
    with jsonlines.open(str(path)) as reader:
        for data in reader:
            data_list.append(data)
    return data_list


def save_as_jsonlines(
    data: List[Dict],
    path: str | Path,
    parents: bool = True,
    exist_ok: bool = True,
) -> None:
    path = Path(path)
    path.parent.mkdir(parents=parents, exist_ok=exist_ok)
    with jsonlines.open(str(path), mode="w") as writer:
        for datum in data:
            writer.write(datum)
    return


def load_json(path: str | Path) -> Dict:
    with open(str(path)) as f:
        data = json.load(f)
    return data  # type: ignore


def save_formatted_json(
    data: Dict | List | str,
    path: str | Path,
    parents: bool = True,
    exist_ok: bool = True,
) -> None:
    """
    Saves a dictionary or a list which is JSON serializable to a formatted JSON
    (UTF-8, 4 space indent).
    Paramters:
        data: A dictionary or a list which is JSON serializable. JSON data as string
            can also be input.
        path: Path to save the input `data` to.
        parents: Determines whether to make parent directories of the output file.
            Will be input to `pathlib.Path.mkdir` method.
        exist_ok: Determines whether to make parent directory if it exists already.
            Will be input to `pathlib.Path.mkdir` method.
    """
    path = Path(path)
    path.parent.mkdir(parents=parents, exist_ok=exist_ok)
    if isinstance(data, str):
        data = json.loads(data)
    with path.open(mode="w", encoding="utf-8") as fout:
        json.dump(data, fout, ensure_ascii=False, indent=4, separators=(",", ": "))
    return None


def save_pickle(
    data: Dict | List | str,
    path: str | Path,
    parents: bool = True,
    exist_ok: bool = True,
) -> None:
    path = Path(path)
    path.parent.mkdir(parents=parents, exist_ok=exist_ok)
    with path.open(mode="wb") as fout:
        pickle.dump(data, fout)
    return None


def load_pickle(
    path: str | Path,
) -> Dict | List | str:
    path = Path(path)
    with path.open(mode="rb") as fin:
        data = pickle.load(fin)
    return data  # type: ignore



import os
from pathlib import Path

DATA_DIR = Path(os.getenv("DATA_DIR", Path(__file__).parents[1] / "data"))
PROJECT_DIR = Path(os.getenv("PROJECT_DIR", Path(__file__).parents[1]))
OUTPUT_DIR = Path(os.getenv("OUTPUT_DIR", Path(__file__).parents[1] / "output"))



import logging

import torch
from transformers import GPT2Config, GPT2Model

from rlearn.sports.soccer.modules.sequence_decoder.sequence_decoder import SequenceDecoder

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


@SequenceDecoder.register("huggingface_gpt2_decoder")
class HuggingfaceGPT2Decoder(SequenceDecoder):
    """
    Huggingface GPT2 Decoder

    Args:
        input_dim: input dimension
        num_layers: number of layers
        feedforward_hidden_dim: feedforward hidden dimension
        num_attention_head: number of attention heads
        activation: activation function
        n_positions: number of positions
        pretrained_model_name_or_path: pretrained model name or path
    """

    def __init__(
        self,
        input_dim: int,
        num_layers: int,
        feedforward_hidden_dim: int,
        num_attention_head: int,
        activation: str = "gelu_new",
        n_positions: int = 8192,
        pretrained_model_name_or_path: str | None = None,
    ) -> None:
        super().__init__()
        if pretrained_model_name_or_path is not None:
            logger.info(f"Loading pretrained model from {pretrained_model_name_or_path}")
            self.model = GPT2Model.from_pretrained(pretrained_model_name_or_path)
            self.config = self.model.config
            self.input_dim = self.config.n_embd
            self.num_layers = self.config.n_layer
            self.feedforward_hidden_dim = self.config.n_inner
            self.num_attention_head = self.config.n_head
            self.activation = self.config.activation_function
            self.n_positions = self.config.n_positions
        else:
            logger.info("Building model from scratch")
            self.config = GPT2Config(
                n_positions=n_positions,
                n_embd=input_dim,
                n_layer=num_layers,
                n_head=num_attention_head,
                n_inner=feedforward_hidden_dim,
                activation_function=activation,
                output_hidden_states=True,
            )
            self.model = GPT2Model(self.config)
            self.input_dim = input_dim
            self.num_layers = num_layers
            self.feedforward_hidden_dim = feedforward_hidden_dim
            self.num_attention_head = num_attention_head
            self.activation = activation
            self.n_positions = n_positions

    def get_input_dim(self) -> int:
        return self.input_dim

    def get_output_dim(self) -> int:
        return self.input_dim

    def is_bidirectional(self) -> bool:
        return False

    def forward(self, inputs: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        """
        Args:
            inputs: (batch_size, seq_len, input_dim)
            mask: (batch_size, seq_len)
        Returns:
            output: (batch_size, seq_len, input_dim)
        """
        output = self.model(inputs_embeds=inputs, attention_mask=mask)
        return output.last_hidden_state



from typing import Any, Dict, List, Tuple, cast

import matplotlib.pyplot as plt
import torch
import torch.nn as nn
import torch.nn.functional as F

from rlearn.sports.soccer.models.q_model_base import QModelBase
from rlearn.sports.soccer.modules.optimizer import LRScheduler, Optimizer
from rlearn.sports.soccer.modules.seq2seq_encoder import Seq2SeqEncoder
from rlearn.sports.soccer.modules.token_embedder import TokenEmbedder
from rlearn.sports.soccer.torch.metrics.classification import get_classification_full_metrics

class Prediction:
    def __init__(self, q_values: torch.Tensor) -> None:
        self.q_values = q_values

@QModelBase.register("sarsa_attacker")
class AttacckerSARSAModel(QModelBase):
    def __init__(
        self,
        observation_dim: int,
        sequence_encoder: Dict[str, Any],
        optimizer: Dict[str, Any] | None = None,
        scheduler: Dict[str, Any] | None = None,
        vocab_size: int = 16,
        pad_token_id: int = 15,
        gamma: float = 1.0,
        lambda_: float = 0.0,
        lambda2_: float = 0.0,
        class_weights: torch.Tensor | None = None,
    ):
        super().__init__()
        self.observation_dim = observation_dim
        self.vocab_size = vocab_size
        self._optimizer_config = optimizer
        self._scheduler_config = scheduler
        self.pad_token_id = pad_token_id
        self.gamma = gamma
        self.lambda_ = lambda_
        self.lambda2_ = lambda2_
        self.class_weights = class_weights
        self.class_weights = torch.tensor(class_weights) if class_weights is not None else None

        self.encoder = Seq2SeqEncoder.from_params(sequence_encoder)
        self.fc1 = nn.Linear(self.observation_dim, self.encoder.get_input_dim())
        self.fc2 = nn.Linear(self.encoder.get_output_dim(), self.vocab_size)
        print(f"vocab_size: {self.vocab_size}")
        print(f"pad_token_id: {self.pad_token_id}")
        print(f"self.input_dim: {self.encoder.get_input_dim()}")
        print(f"self.output_dim: {self.encoder.get_output_dim()}")
        print(f"self.observation_dim: {self.observation_dim}")

        # evaluation
        self.train_metrics = get_classification_full_metrics(
            num_classes=self.vocab_size,
            average_method="macro",
            ignore_index=self.pad_token_id,
            prefix="train_",
        )
        self.val_metrics = get_classification_full_metrics(
            num_classes=self.vocab_size,
            average_method="macro",
            ignore_index=self.pad_token_id,
            prefix="val_",
        )
        self.test_metrics = get_classification_full_metrics(
            num_classes=self.vocab_size,
            average_method="macro",
            ignore_index=self.pad_token_id,
            prefix="test_",
        )

    def forward(self, batch: Dict[str, torch.Tensor]) -> torch.Tensor:
        # observation: (batch_size, seq_len, obs_dim)
        x = F.relu(self.fc1(batch['observation']))  # (batch_size, input_length, encoder_input_dim)
        out = self.encoder(x)
        q_values = self.fc2(out)  # (batch_size, input_length, vocab_size)

        # if q_values contains NaN values, print the indices of the NaN values and the values themselves
        if torch.isnan(q_values).any():
            print("q_values contains NaN values")
            nan_indices = torch.nonzero(torch.isnan(q_values)).squeeze()
            for idx in nan_indices:
                idx_tuple = tuple(idx.tolist())
                print(f"NaN at index: {idx_tuple}")
                print(f"q_values[{idx_tuple}] = {q_values[idx_tuple]}")
                print(f"batch['observation'][{idx_tuple[:-1]}] = {batch['observation'][idx_tuple[:-1]]}")
                print(f"batch['action'][{idx_tuple[:-1]}] = {batch['action'][idx_tuple[:-1]]}")
                print(f"batch['reward'][{idx_tuple[:-1]}] = {batch['reward'][idx_tuple[:-1]]}")
                print(f"batch['mask'][{idx_tuple[:-1]}] = {batch['mask'][idx_tuple[:-1]]}")
        return q_values

    def training_step(self, batch: Dict[str, torch.Tensor], batch_idx: int) -> torch.Tensor:
        # avail_actions = []
        mask = batch['mask']  # (batch_size, input_length)
        reward = batch['reward']  # (batch_size, input_length)
        action = batch['action']  # (batch_size, input_length)

        q_values = self.forward(batch)  # (batch_size, input_length, vocab_size)
        pred_q_values = q_values.argmax(dim=2)  # (batch_size, input_length)
        td_loss = (reward[:, 1:] + self.gamma * pred_q_values[:, 1:] - pred_q_values[:, :-1]) ** 2
        td_loss = (td_loss * mask[:, 1:]).sum() / mask[:, 1:].sum()
        l1_loss = sum([torch.norm(param, p=1) for param in self.parameters() if param.requires_grad])
        self.log("train_td_loss", td_loss)
        self.log("train_l1_loss", l1_loss)

        action_loss = F.cross_entropy(
            input=q_values.reshape(-1, self.vocab_size),
            target=action.reshape(-1),
            reduction="mean",
            ignore_index=self.pad_token_id,
            weight=self.class_weights.to(q_values.device) if self.class_weights is not None else None,
        )
        
        self.train_metrics(
            q_values.reshape(-1, self.vocab_size),
            action.reshape(-1),
        )
        self.log("train_action_loss", action_loss)
        self.log_dict(self.train_metrics)

        total_loss = td_loss + self.lambda_ * l1_loss + self.lambda2_ * action_loss
        self.log("train_total_loss", total_loss)

        return total_loss
    
    def validation_step(self, batch: Dict[str, torch.Tensor], batch_idx: int) -> torch.Tensor:
        mask = batch['mask']  # (batch_size, input_length)
        reward = batch['reward']  # (batch_size, input_length)
        action = batch['action']  # (batch_size, input_length)

        q_values = self.forward(batch)  # (batch_size, input_length, vocab_size)
        pred_q_values = q_values.argmax(dim=2)  # (batch_size, input_length)
        print(f"pred q values: {pred_q_values}")
        td_loss = (reward[:, 1:] + self.gamma * pred_q_values[:, 1:] - pred_q_values[:, :-1]) ** 2
        td_loss = (td_loss * mask[:, 1:]).sum() / mask[:, 1:].sum()
        # L1 loss for parameter regularization
        l1_loss = sum([torch.norm(param, p=1) for param in self.parameters() if param.requires_grad])
        self.log("val_td_loss", td_loss)
        self.log("val_l1_loss", l1_loss)

        action_loss = F.cross_entropy(
            input=q_values.reshape(-1, self.vocab_size),
            target=action.reshape(-1),
            reduction="mean",
            ignore_index=self.pad_token_id,
            weight=self.class_weights.to(q_values.device) if self.class_weights is not None else None,
        )
        self.val_metrics(
            q_values.reshape(-1, self.vocab_size),
            action.reshape(-1),
        )
        self.log("val_action_loss", action_loss)
        self.log_dict(self.val_metrics)

        total_loss = td_loss + self.lambda_ * l1_loss + self.lambda2_ * action_loss
        self.log("val_loss", total_loss)

        # log prediction count as a histogram
        pred_actions = q_values.argmax(dim=2)[batch['mask']]
        class_counts = torch.bincount(pred_actions, minlength=self.vocab_size).cpu().numpy()
        class_ratios = class_counts / class_counts.sum()
        fig, ax = plt.subplots()
        ax.bar(range(self.vocab_size), class_ratios)
        ax.set_xticks(range(self.vocab_size))
        ax.set_xlabel('Classes')
        ax.set_ylabel('Predicted Ratio')
        self.logger.experiment.add_figure("Predicted Class Ratios (validation)", fig, self.current_epoch)

        # log gold count as a histogram
        gold_actions = batch['action'][batch['mask']]
        class_counts = torch.bincount(gold_actions.flatten(), minlength=self.vocab_size).cpu().numpy()
        class_ratios = class_counts / class_counts.sum()
        fig, ax = plt.subplots()
        ax.bar(range(self.vocab_size), class_ratios)
        ax.set_xticks(range(self.vocab_size))
        ax.set_xlabel('Classes')
        ax.set_ylabel('Gold Ratio')
        self.logger.experiment.add_figure("Gold Class Ratios (validation)", fig, self.current_epoch)

        return total_loss

    def test_step(self, batch: Dict[str, torch.Tensor], batch_idx: int) -> torch.Tensor:
        mask = batch['mask']  # (batch_size, input_length)
        reward = batch['reward']  # (batch_size, input_length)
        action = batch['action']  # (batch_size, input_length)

        q_values = self.forward(batch)  # (batch_size, input_length, vocab_size)
        self.check_for_nan_inf(q_values, "q_values")

        pred_q_values = q_values.argmax(dim=2)  # (batch_size, input_length)
        print(f"pred q values: {pred_q_values}")
        self.check_for_nan_inf(pred_q_values, "pred_q_values")

        td_loss = (reward[:, 1:] + self.gamma * pred_q_values[:, 1:] - pred_q_values[:, :-1]) ** 2
        self.check_for_nan_inf(td_loss, "td_loss before mask")
        td_loss = (td_loss * mask[:, 1:]).sum() / mask[:, 1:].sum()
        self.check_for_nan_inf(td_loss, "td_loss after mask")

        l1_loss = sum([torch.norm(param, p=1) for param in self.parameters() if param.requires_grad])
        self.check_for_nan_inf(l1_loss, "l1_loss")
        self.log("test_td_loss", td_loss)
        self.log("test_l1_loss", l1_loss)

        action_loss = F.cross_entropy(
            input=q_values.reshape(-1, self.vocab_size),
            target=action.reshape(-1),
            reduction="mean",
            ignore_index=self.pad_token_id,
            weight=self.class_weights.to(q_values.device) if self.class_weights is not None else None,
        )
        self.check_for_nan_inf(action_loss, "action_loss")

        # import pdb; pdb.set_trace()

        self.test_metrics(
            q_values.reshape(-1, self.vocab_size),
            action.reshape(-1),
        )
        self.log("test_action_loss", action_loss)
        self.log_dict(self.test_metrics)

        total_loss = td_loss + self.lambda_ * l1_loss + self.lambda2_ * action_loss
        self.check_for_nan_inf(total_loss, "total_loss")
        self.log("test_loss", total_loss)

        # log prediction count as a histogram
        pred_actions = q_values.argmax(dim=2)[batch['mask']]
        class_counts = torch.bincount(pred_actions, minlength=self.vocab_size).cpu().numpy()
        class_ratios = class_counts / class_counts.sum()
        fig, ax = plt.subplots()
        ax.bar(range(self.vocab_size), class_ratios)
        ax.set_xticks(range(self.vocab_size))
        ax.set_xlabel('Classes')
        ax.set_ylabel('Predicted Ratio')
        self.logger.experiment.add_figure("Predicted Class Ratios (test)", fig, self.current_epoch)

        # log gold count as a histogram
        gold_actions = batch['action'][batch['mask']]
        class_counts = torch.bincount(gold_actions, minlength=self.vocab_size).cpu().numpy()
        class_ratios = class_counts / class_counts.sum()
        fig, ax = plt.subplots()
        ax.bar(range(self.vocab_size), class_ratios)
        ax.set_xticks(range(self.vocab_size))
        ax.set_xlabel('Classes')
        ax.set_ylabel('Gold Ratio')
        self.logger.experiment.add_figure("Gold Class Ratios (test)", fig, self.current_epoch)

        return total_loss


    def check_for_nan_inf(self, tensor, name):
        if torch.isnan(tensor).any():
            print(f"{name} contains NaN values")
        if torch.isinf(tensor).any():
            print(f"{name} contains Inf values")

    def configure_optimizers(self) -> Tuple[List[Optimizer], List[LRScheduler]] | Optimizer:
        self._optimizer = Optimizer.from_params(params_=self._optimizer_config, params=self.parameters())
        if self._scheduler_config is not None:
            self._scheduler = LRScheduler.from_params(params_=self._scheduler_config, optimizer=self._optimizer)
            return [self._optimizer], [self._scheduler]
        return self._optimizer



from typing import List
import numpy as np
import math

import torch
from pydantic import BaseModel, ConfigDict, field_validator, model_validator


class Position(BaseModel):
    x: float
    y: float

    def to_dict(self) -> dict:
        return {"x": self.x, "y": self.y}

    @classmethod
    def from_dict(cls, d: dict) -> "Position":
        return cls(x=d["x"], y=d["y"])

    def distance_to(self, other: "Position") -> float:
        return ((self.x - other.x) ** 2 + (self.y - other.y) ** 2) ** 0.5

    def angle_to(self, other: "Position") -> float:
        return math.atan2(other.y - self.y, other.x - self.x)


class Velocity(BaseModel):
    x: float
    y: float

    def to_dict(self) -> dict:
        return {"x": self.x, "y": self.y}

    @classmethod
    def from_dict(cls, d: dict) -> "Velocity":
        return cls(x=d["x"], y=d["y"])


class Player(BaseModel):
    index: int
    team_name: str
    player_name: str
    player_id: int
    player_role: str
    position: Position
    velocity: Velocity
    action: str
    action_probs: List[float] | None = None

    def to_dict(self) -> dict:
        return {
            "index": self.index,
            "team_name": self.team_name,
            "player_name": self.player_name,
            "player_id": self.player_id,
            "player_role": self.player_role,
            "position": self.position.to_dict(),
            "velocity": self.velocity.to_dict(),
            "action": self.action,
            "action_probs": self.action_probs or None,
        }

    @classmethod
    def from_dict(cls, d: dict) -> "Player":
        return cls(
            index=d["index"],
            team_name=d["team_name"],
            player_name=d["player_name"],
            player_id=d["player_id"],
            player_role=d["player_role"],
            position=Position.from_dict(d["position"]),
            velocity=Velocity.from_dict(d["velocity"]),
            action=d["action"],
            action_probs=d["action_probs"] if "action_probs" in d else None,
        )


class Ball(BaseModel):
    position: Position
    velocity: Velocity

    def to_dict(self) -> dict:
        return {
            "position": self.position.to_dict(),
            "velocity": self.velocity.to_dict(),
        }

    @classmethod
    def from_dict(cls, d: dict) -> "Ball":
        return cls(
            position=Position.from_dict(d["position"]),
            velocity=Velocity.from_dict(d["velocity"]),
        )

class State(BaseModel):
    ball: Ball
    players: List[Player]
    attack_players: List[Player]
    defense_players: List[Player]

    @field_validator("attack_players", "defense_players")
    @classmethod
    def players_must_be_list_of_players(cls, v):  # type: ignore
        if not isinstance(v, list):
            raise TypeError("players must be a list")
        for player in v:
            if not isinstance(player, Player):
                raise TypeError("players must be a list of Player")
        return v

    def to_dict(self) -> dict:
        return {
            "ball": self.ball.to_dict(),
            "players": [player.to_dict() for player in self.players],
            "attack_players": [player.to_dict() for player in self.attack_players],
            "defense_players": [player.to_dict() for player in self.defense_players],
        }

    @classmethod
    def from_dict(cls, d: dict) -> "State":
        return cls(
            ball=Ball.from_dict(d["ball"]),
            players=[Player.from_dict(player) for player in d["players"]],
            attack_players=[Player.from_dict(player) for player in d["attack_players"]],
            defense_players=[Player.from_dict(player) for player in d["defense_players"]],
        )
    

class Observation(BaseModel):
    ball: Ball
    players: List[Player]  # without ego_player
    ego_player: Player

    @classmethod
    def from_state(cls, state: State, ego_player: Player) -> "Observation":
        raise NotImplementedError

    def to_tensor(self) -> torch.Tensor:
        raise NotImplementedError

    @classmethod
    def from_dict(cls, data):
        raise NotImplementedError

    def to_dict(self):
        raise NotImplementedError

    @field_validator("players")
    @classmethod
    def players_must_be_list_of_players(cls, v):  # type: ignore
        if not isinstance(v, list):
            raise TypeError("players must be a list")
        for player in v:
            if not isinstance(player, Player):
                raise TypeError("players must be a list of Player")
        return v


class SimpleObservation(Observation):
    @classmethod
    def from_state(cls, state: State, ego_player: Player) -> "SimpleObservation":
        ego_position = ego_player.position
        ego_player_id = ego_player.player_id
        players = [
            Player(
                index=player.index,
                team_name=player.team_name,
                player_name=player.player_name,
                player_id=player.player_id,
                player_role=player.player_role,
                position=Position(x=player.position.x - ego_position.x, y=player.position.y - ego_position.y),
                velocity=player.velocity,
                action=player.action,
                action_probs=player.action_probs,
            )
            for player in state.players
            if player.player_id != ego_player_id
        ]
        ball = Ball(
            position=state.ball.position,
            velocity=state.ball.velocity,
        )
        return cls(ball=ball, players=players, ego_player=ego_player)

    def to_tensor(self) -> torch.Tensor:
        data = []
        for player in self.players:
            data.extend([player.position.x, player.position.y, player.velocity.x, player.velocity.y])
        data.extend([self.ball.position.x, self.ball.position.y, self.ball.velocity.x, self.ball.velocity.y])
        data.extend([self.ego_player.position.x, self.ego_player.position.y, self.ego_player.velocity.x, self.ego_player.velocity.y])
        return torch.tensor(data)

    @property
    def dimension(self) -> int:
        return len(self.to_tensor())

    def to_dict(self):
        return {
            "ball": self.ball.to_dict(),
            "players": [player.to_dict() for player in self.players],
            "ego_player": self.ego_player.to_dict(),
        }

    @classmethod
    def from_dict(cls, data):
        return cls(
            ball=Ball.from_dict(data["ball"]),
            players=[Player.from_dict(player) for player in data["players"]],
            ego_player=Player.from_dict(data["ego_player"]),
        )


class SimpleObservationAction(BaseModel):
    player: Player
    observation: SimpleObservation
    action: str
    reward: float

    def to_dict(self):
        return {
            "player": self.player.to_dict(),
            "observation": self.observation.to_dict(),
            "action": self.action,
            "reward": self.reward,
        }

    @classmethod
    def from_dict(cls, data):
        return cls(
            player=Player.from_dict(data["player"]),
            observation=SimpleObservation.from_dict(data["observation"]),
            action=data["action"],
            reward=data["reward"],
        )


class SimpleObservationActionSequence(BaseModel):
    game_id: str
    half: str
    sequence_id: int
    team_name_attack: str
    team_name_defense: str
    sequence: List[SimpleObservationAction]

    def to_dict(self):
        return {
            "game_id": self.game_id,
            "half": self.half,
            "sequence_id": self.sequence_id,
            "team_name_attack": self.team_name_attack,
            "team_name_defense": self.team_name_defense,
            "sequence": [obs_action.to_dict() for obs_action in self.sequence],
        }

    @classmethod
    def from_dict(cls, data):
        return cls(
            game_id=data["game_id"],
            half=data["half"],
            sequence_id=data["sequence_id"],
            team_name_attack=data["team_name_attack"],
            team_name_defense=data["team_name_defense"],
            sequence=[SimpleObservationAction.from_dict(obs_action) for obs_action in data["sequence"]],
        )

class Event(BaseModel):
    state: State
    action: List[str] | None = None
    reward: float

    @model_validator(mode="after")
    def set_and_validate_action(self) -> "Event":
        if self.action is None:
            self.action = [player.action for player in self.state.players]
        for action in self.action:
            if not isinstance(action, str):
                raise TypeError("action must be a list of int")
        return self

    def to_dict(self) -> dict:
        return {
            "state": self.state.to_dict(),
            "action": self.action,
            "reward": self.reward,
        }

    @classmethod
    def from_dict(cls, d: dict) -> "Event":
        return cls(
            state=State.from_dict(d["state"]),
            action=d["action"],
            reward=d["reward"],
        )


class Events(BaseModel):
    game_id: str
    half: str
    sequence_id: int
    sequence_start_frame: str
    sequence_end_frame: str
    team_name_attack: str
    team_name_defense: str
    events: List[Event]

    @field_validator("events")
    @classmethod
    def events_must_be_list_of_events(cls, v):  # type: ignore
        if not isinstance(v, list):
            raise TypeError("events must be a list")
        for event in v:
            if not isinstance(event, Event):
                raise TypeError("events must be a list of Event")
        return v

    def to_dict(self) -> dict:
        return {
            "game_id": self.game_id,
            "half": self.half,
            "sequence_id": self.sequence_id,
            "sequence_start_frame": self.sequence_start_frame,
            "sequence_end_frame": self.sequence_end_frame,
            "team_name_attack": self.team_name_attack,
            "team_name_defense": self.team_name_defense,
            "events": [event.to_dict() for event in self.events],
        }

    @classmethod
    def from_dict(cls, d: dict) -> "Events":
        return cls(
            game_id=d["game_id"],
            half=d["half"],
            sequence_id=d["sequence_id"],
            sequence_start_frame=d["sequence_start_frame"],
            sequence_end_frame=d["sequence_end_frame"],
            team_name_attack=d["team_name_attack"],
            team_name_defense=d["team_name_defense"],
            events=[Event.from_dict(event) for event in d["events"]],
        )

class ObservationactionInstance(BaseModel):
    observation: torch.Tensor  # (events_len, n_agents, obs_dim)
    action: torch.Tensor  # (events_len, n_agents)
    reward: torch.Tensor  # (events_len, n_agents)
    model_config = ConfigDict(arbitrary_types_allowed=True)


class ObservationactionBatch(BaseModel):
    observation: torch.Tensor  # (batch_size, max_events_len, n_agents, obs_dim)
    action: torch.Tensor  # (batch_size, max_events_len, n_agents)
    reward: torch.Tensor  # (batch_size, max_events_len, n_agents)
    mask: torch.Tensor  # (batch_size, max_events_len, n_agents)
    model_config = ConfigDict(arbitrary_types_allowed=True)


class Prediction(BaseModel):
    q_values: torch.Tensor  # (batch_size, max_events_len, n_agents, action_dim)
    model_config = ConfigDict(arbitrary_types_allowed=True)


class ObservationactionForLMInstance(BaseModel):
    sequence: List[int]
    action_mask: List[int]  # 1 if action, else 0


class ObservationActionForLMBatch(BaseModel):
    sequence: torch.Tensor  # (batch_size, max_seq_len )
    mask: torch.Tensor  # (batch_size, max_seq_len)
    action_mask: torch.Tensor  # (batch_size, max_seq_len)
    model_config = ConfigDict(arbitrary_types_allowed=True)


class ActionPrediction(BaseModel):
    q_values: torch.Tensor  # (batch_size, max_events_len, action_dim)
    model_config = ConfigDict(arbitrary_types_allowed=True)



import logging
import re
from pathlib import Path, PosixPath
from typing import Any, Dict
from copy import deepcopy
import time
from tqdm import tqdm
import pandas as pd
from datasets import load_dataset, load_from_disk
from sklearn.model_selection import train_test_split
import pytorch_lightning as pl
import torch
from lightning_lite.utilities.seed import seed_everything


from ..dataclass import Events, SimpleObservation, SimpleObservationAction, SimpleObservationActionSequence
from ..utils.file_utils import load_json, save_formatted_json
from ..env import OUTPUT_DIR, PROJECT_DIR
from ..models.q_model_base import QModelBase
from ..modules.datamodule import DataModule
from ..class_weight.class_weight import ClassWeightBase
from ..application.q_values_movie import create_movie


logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

class rlearn_model_soccer:
    def __init__(self, model_name=None, config=None, 
                seed=42, num_process=4, input_path=None, output_path=None,
                ):
        self.model_name = model_name   
        self.config = config
        self.seed = seed
        self.num_process = num_process
        self.input_path = input_path
        self.output_path = output_path

        
    def train_test_split(self):
        # Load data into a Dataset
        game_ids = [str(p.name) for p in Path(self.input_path).glob("*") if re.match(r"\d{7}", p.name)]
        train_game_ids, test_val_game_ids = train_test_split(game_ids, test_size=0.5, random_state=self.seed)
        test_game_ids, val_game_ids = train_test_split(test_val_game_ids, test_size=0.1, random_state=self.seed)

        train_dataset = load_dataset(
            "json",
            data_files=[str(Path(self.input_path) / f"{game_id}" / "events.jsonl") for game_id in train_game_ids],
            split="train",
            num_proc=self.num_process,
        )
        valid_dataset = load_dataset(
            "json",
            data_files=[str(Path(self.input_path) / f"{game_id}" / "events.jsonl") for game_id in val_game_ids],
            split="train",
            num_proc=self.num_process,
        )
        test_dataset = load_dataset(
            "json",
            data_files=[str(Path(self.input_path) / f"{game_id}" / "events.jsonl") for game_id in test_game_ids],
            split="train",
            num_proc=self.num_process,
        )

        output_dir = Path(self.output_path)
        # Set output directory
        if output_dir is None:
            output_dir = self.input_path
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

        # Save the splits
        for split_name, split_dataset in zip(
            ["train", "validation", "test"],
            [train_dataset, valid_dataset, test_dataset],
        ):
            split_dataset.save_to_disk(output_dir / split_name)
        # for debugging
        train_dataset.select(range(5)).save_to_disk(output_dir / "mini")

        logging.info(f"Data splits saved in {output_dir}:")
        logging.info(f"Train: {len(train_dataset)}")
        logging.info(f"Validation: {len(valid_dataset)}")
        logging.info(f"Test: {len(test_dataset)}")

    
    def events2attacker_simple_observation_action_sequence(
            self, examples, min_frame_len_threshold: int=30, max_frame_len_threshold: int=500
        )->Dict[str, Any]:
        events_list = [Events.from_dict(dict(zip(examples, v))) for v in zip(*examples.values())]
        for events in events_list:
            assert (
                min_frame_len_threshold <= len(events.events) <= max_frame_len_threshold
            ), f"len(events.events): {len(events.events)}"
        attacker_observation_action_sequence = []
        for events in events_list:
            valid_attack_player_ids = [
                player.player_id
                for player in events.events[0].state.players
                if player.player_role != "GK" and player.player_id > 0 and player.team_name == events.team_name_attack
            ]
            for target_player_id in valid_attack_player_ids:
                attacker_observation_action_sequence_in_event = []
                for event in events.events:
                    try:
                        target_player = [
                            player for player in event.state.attack_players if player.player_id == target_player_id
                        ][0]
                    except IndexError:
                        logger.warning(
                            f"target_player_id: {target_player_id} not found in game_id: {events.game_id} half: {events.half} seq_id: {events.sequence_id}"
                        )
                        continue
                    observation = SimpleObservation.from_state(event.state, target_player)
                    action = target_player.action
                    observation_action = SimpleObservationAction(
                        player=target_player, observation=observation, action=action, reward=event.reward
                    )
                    attacker_observation_action_sequence_in_event.append(observation_action)
                attacker_observation_action_sequence.append(attacker_observation_action_sequence_in_event)
        attacker_observation_action_sequence = [
            SimpleObservationActionSequence(
                game_id=events.game_id,
                half=events.half,
                sequence_id=events.sequence_id,
                team_name_attack=events.team_name_attack,
                team_name_defense=events.team_name_defense,
                sequence=observation_action_sequence,
            ).to_dict()
            for _, observation_action_sequence in enumerate(attacker_observation_action_sequence)
        ]
        return {
            key: [item[key] for item in attacker_observation_action_sequence]
            for key in attacker_observation_action_sequence[0].keys()
        }


    def preprocess_observation(self, batch_size):
        logging.info(f"input_path: {self.input_path}")
        start = time.time()
        config = load_json(self.config)
        dataset = load_from_disk(str(self.input_path))
        logger.info("Length of dataset: {}".format(len(dataset)))
        dataset = dataset.map(
            self.events2attacker_simple_observation_action_sequence,
            batched=True,
            remove_columns=dataset.column_names,
            num_proc=self.num_process,
            batch_size=batch_size,
            fn_kwargs={
                "min_frame_len_threshold": config["min_frame_len_threshold"],
                "max_frame_len_threshold": config["max_frame_len_threshold"],
            },
        )
        logger.info("Length of dataset after processing: {}".format(len(dataset)))
        dataset.save_to_disk(str(self.output_path))
        logging.info(f"output_path: {self.output_path} (elapsed: {time.time() - start:.2f} sec)")


    def train(self, exp_name, run_name, accelerator, devices, strategy):
        seed_everything(self.seed)
        exp_config = load_json(self.config)
        config_copy = deepcopy(exp_config)
        logger.info(f"exp_config: {exp_config}")
        output_dir = OUTPUT_DIR / exp_name / run_name
        output_dir.mkdir(exist_ok=True, parents=True)

        logger.info("loading dataset...")
        train_dataset = load_from_disk(Path(exp_config['dataset']['train_filename']).resolve())
        valid_dataset = load_from_disk(Path(exp_config['dataset']['valid_filename']).resolve())
        test_dataset = load_from_disk(Path(exp_config['dataset']['test_filename']).resolve())
        logger.info("Preprocessing dataset...")
        start = time.time()
        train_dataset = DataModule.by_name(exp_config['datamodule']['type']).preprocess_data(
            train_dataset, **exp_config['dataset']['preprocess_config']
        )
        valid_dataset = DataModule.by_name(exp_config['datamodule']['type']).preprocess_data(
            valid_dataset, **exp_config['dataset']['preprocess_config']
        )
        test_dataset = DataModule.by_name(exp_config['datamodule']['type']).preprocess_data(
            test_dataset, **exp_config['dataset']['preprocess_config']
        )
        logger.info(f"Preprocessing dataset is done. {time.time() - start} sec")
        logger.info(f"Train dataset size: {len(train_dataset)}")
        logger.info(f"Valid dataset size: {len(valid_dataset)}")
        logger.info(f"Test dataset size: {len(test_dataset)}")

        datamodule = DataModule.from_params(
            params_=exp_config["datamodule"],
            train_dataset=train_dataset,
            valid_dataset=valid_dataset,
            test_dataset=test_dataset,
        )
        # count tokens and calculate class weights (the inverse of the frequency of each class)
        # cache the class weights so that we need not calculate them every time
        if "class_weight_fn" in exp_config:
            logger.info("Prepare class weights...")
            start = time.time()
            class_weight_fn = ClassWeightBase.from_params(exp_config["class_weight_fn"])
            if class_weight_fn.class_weights is not None:
                class_weights = class_weight_fn.class_weights
            else:
                logger.info("Calculating class weights...")
                class_counts = torch.zeros(datamodule.state_action_tokenizer.num_tokens)
                for batch in tqdm(datamodule.train_dataloader(batch_size=512), desc="calculating class weights"):
                    valid_actions = torch.masked_select(batch['action'], batch['mask'].bool())
                    class_counts += torch.bincount(valid_actions, minlength=datamodule.state_action_tokenizer.num_tokens)
                class_weights = class_weight_fn.calculate(class_counts=class_counts)
            assert class_weights.shape[0] == datamodule.state_action_tokenizer.num_tokens , f"Class weights shape mismatch: {class_weights.shape[0]} != {datamodule.state_action_tokenizer.num_tokens}"
            logger.info(f"Prepare class weights is done. {time.time() - start} sec")
        else:
            class_weights = None
            

        tensorboard_logger = pl.loggers.TensorBoardLogger(
            save_dir=str(output_dir / "tensorboard_logs"),
            name=run_name,
        )
        mlflow_logger = pl.loggers.MLFlowLogger(
            experiment_name=exp_name,
            run_name=run_name,
            save_dir=str(PROJECT_DIR / "mlruns"),
        )
        checkpoint_callback = pl.callbacks.ModelCheckpoint(
            dirpath=output_dir / "checkpoints",
            monitor="val_loss",
            mode="min",
            save_top_k=1,
        )
        early_stopping_callback = pl.callbacks.EarlyStopping(
            monitor="val_loss",
            patience=5,
            mode="min",
            min_delta=0.01,
        )
        trainer = pl.Trainer(
            max_epochs=exp_config["max_epochs"],
            logger=[tensorboard_logger, mlflow_logger],
            callbacks=[checkpoint_callback, early_stopping_callback],
            accelerator=accelerator,
            devices=devices,
            strategy=strategy,
            deterministic=True,
            val_check_interval=exp_config["val_check_interval"] if "val_check_interval" in exp_config else None,
            detect_anomaly=True,
            accumulate_grad_batches=exp_config["accumulate_grad_batches"]
            if "accumulate_grad_batches" in exp_config
            else 1,
            gradient_clip_val=None
        )

        params_ = {
            "type": exp_config["model"]["type"],
            "observation_dim":  exp_config["model"]["observation_dim"],
            "sequence_encoder": exp_config['model']['sequence_encoder'],
            "optimizer": exp_config['model']['optimizer'],
            "vocab_size": datamodule.state_action_tokenizer.num_tokens,
            "pad_token_id": datamodule.state_action_tokenizer.encode("[PAD]"),
            "gamma": exp_config["model"]["gamma"],
            "lambda_": exp_config["model"]["lambda_"],
            "lambda2_": exp_config["model"]["lambda2_"],
            "class_weights": class_weights,
        }
        params_["class_weights"] = params_["class_weights"].tolist()

        print(f"params_: {params_}")
        print("class_weights:", params_['class_weights'])
        
        model = QModelBase.from_params(
            params_=params_
        )

        trainer.fit(model=model, datamodule=datamodule)
        trainer.test(model=model, dataloaders=datamodule.test_dataloader())
        save_formatted_json(config_copy, output_dir / "config.json")


    def visualize_data(self, model_name, checkpoint_path, match_id, sequence_id):
        exp_config_path = os.getcwd()+f"/tests/config/{model_name}.json"
        exp_config = load_json(exp_config_path)
        test_file_path = Path(os.getcwd()+ "/" + exp_config['dataset']['test_filename'])
        test_dataset = load_from_disk(test_file_path)
        test_dataset = DataModule.by_name(exp_config['datamodule']['type']).preprocess_data(
            test_dataset, **exp_config['dataset']['preprocess_config']
        )


        # unique game_ids
        game_ids = set(test_dataset['game_id'])

        for game_id_to_find in game_ids:
            indices = [i for i, game_id in enumerate(test_dataset['game_id']) if game_id == game_id_to_find]
            sequence_ids = [test_dataset['sequence_id'][i] for i in indices]
            unique_sequence_ids = set(sequence_ids)

            print(f"Unique sequence_ids for game_id {game_id_to_find}: {unique_sequence_ids}")

        # import pdb; pdb.set_trace()

        print(f"start loading {match_id} {sequence_id}")

        datamodule = DataModule.from_params(
            exp_config["datamodule"],
            train_dataset=test_dataset,
            valid_dataset=None,
            test_dataset=None,
        )

        type_ = exp_config["model"]["type"]
        observation_dim = exp_config["model"]["observation_dim"]
        sequence_encoder = exp_config['model']['sequence_encoder']
        optimizer = exp_config['model']['optimizer']
        model = QModelBase.from_params(
            params_={
                "type": type_,
                "observation_dim": observation_dim,
                "sequence_encoder": sequence_encoder,
                "vocab_size": datamodule.state_action_tokenizer.num_tokens,
                "optimizer": optimizer,
                "gamma": exp_config["model"]["gamma"],
                "lambda_": exp_config["model"]["lambda_"],
                "lambda2_": exp_config["model"]["lambda2_"],
            }
        )
        state_dict = torch.load(checkpoint_path, weights_only=False)["state_dict"]
        model.load_state_dict(state_dict)
        model.eval()
        model.to('cuda')

        q_values_list = []

        for data, batch in tqdm(
            zip(datamodule.train_dataset, datamodule.train_dataloader(batch_size=1, shuffle=False)),
            total=len(datamodule.train_dataset),
        ):
            q_values_df = pd.DataFrame(
                index=range(len(data['sequence'])),
                columns=[
                    "game_id",
                    "sequence_id",
                    "team_name",
                    "player_name",
                    "q_value",
                    "action_idx",
                    "q_values_for_actions",
                ]
            )
            q_values_df_path = PROJECT_DIR / f"output/figures/{model_name}/players_q_state/q_values_{match_id}_{sequence_id}.csv"
            q_values_df_path.parent.mkdir(parents=True, exist_ok=True)

            if data['game_id'] == match_id and data['sequence_id'] == sequence_id:
                player = data['sequence'][0]['player']
                q_values = (
                    model(datamodule.transfer_batch_to_device(batch, 'cuda', 0)).squeeze(0).detach().cpu()
                )  # (seq_len, num_actions)
                action_idx = batch['action'].squeeze(0)  # (seq_len, )

                # gather q_values for the actions taken
                q_values_for_actions = q_values.gather(1, action_idx.unsqueeze(1)).squeeze(1).tolist()  # len = seq_len

                for i, _ in enumerate(data['sequence']):
                    q_values_df.loc[i, "game_id"] = data['game_id']
                    q_values_df.loc[i, "sequence_id"] = data['sequence_id']
                    q_values_df.loc[i, "player_name"] = player['player_name']
                    q_values_df.loc[i, "q_value"] = q_values[i, :]
                    q_values_df.loc[i, "action_idx"] = action_idx[i]
                    q_values_df.loc[i, "q_values_for_actions"] = q_values_for_actions[i]

            else:
                continue

            q_values_list.append(q_values_df)

            final_q_values_df = pd.concat(q_values_list, ignore_index=True)
            final_q_values_df.to_csv(q_values_df_path, index=False)

        create_movie(
            q_values_path=q_values_df_path,
            match_id=match_id,
            sequence_id=sequence_id,
        )


if __name__ == '__main__':
    import os
    # split data into train and test
    # rlearn_model_soccer(
    #     input_path=os.getcwd()+'/tests/data/datastadium/',
    #     output_path=os.getcwd()+'/tests/data/datastadium/split/'
    # ).train_test_split()


    # # preprocess observation
    # rlearn_model_soccer(
    #     config=os.getcwd()+'/tests/config/preprocessing_dssports2020.json',
    #     input_path=os.getcwd()+'/tests/data/datastadium/split/mini',
    #     output_path=os.getcwd()+'/tests/data/datastadium_simple_obs_action_seq/split/mini',
    #     num_process=5,
    # ).preprocess_observation(batch_size=64)


    # # train model
    # rlearn_model_soccer(
    #     config=os.getcwd()+'/tests/config/exp_config.json'
    # ).train(
    #     exp_name='sarsa_attacker',
    #     run_name='test',
    #     accelerator="gpu",
    #     devices=1,
    #     strategy="ddp",
    # )


    # visualize data
    rlearn_model_soccer().visualize_data(
        model_name='exp_config',
        checkpoint_path=os.getcwd()+'/rlearn/sports/output/sarsa_attacker/test/checkpoints/epoch=1-step=2.ckpt',
        match_id='2022100106',
        sequence_id=0,
    )


    print('Done')


import logging
import numpy as np

from rlearn.sports.soccer.constant import STOP_THRESHOLD
logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def discretize_direction(velocity_x: float, velocity_y: float) -> str:
    """
    Discretize the direction of the ball/player into 8 directions
        - idle: 0 (when velocity is below STOP_THRESHOLD)
        - right: 1
        - up_right 2
        - up: 3
        - up_left: 4
        - left: 5
        - down_left: 6
        - down: 7
        - down_right: 8
    """

    # if velocity is below threshold, then idle
    if np.sqrt(velocity_x**2 + velocity_y**2) < STOP_THRESHOLD:
        return "idle"

    # calculate angle
    angle = np.arctan2(velocity_y, velocity_x)
    angle = np.rad2deg(angle)
    angle = (angle + 360) % 360

    # discretize angle into 8 directions
    if 22.5 <= angle < 67.5:
        direction = "up_right"
    elif 67.5 <= angle < 112.5:
        direction = "up"
    elif 112.5 <= angle < 157.5:
        direction = "up_left"
    elif 157.5 <= angle < 202.5:
        direction = "left"
    elif 202.5 <= angle < 247.5:
        direction = "down_left"
    elif 247.5 <= angle < 292.5:
        direction = "down"
    elif 292.5 <= angle < 337.5:
        direction = "down_right"
    else:
        direction = "right"

    return direction


from rlearn.sports.soccer.modules.datamodule.datamodule import DataModule  # noqa: F401
from rlearn.sports.soccer.modules.datamodule.jleague_observation_action import (  # noqa: F401
    JLeagueSimpleObservationActionSequenceDataModule,
)
from rlearn.sports.soccer.modules.datamodule.jleague_observation_action_with_eos import (  # noqa: F401
    JLeagueSimpleObservationActionSequenceWithEOSDataModule,
)
from rlearn.sports.soccer.modules.datamodule.jleague_only_attackers import JLeagueRLAttackerDataModule  # noqa: F401


from typing import Literal, Optional

from torchmetrics import Accuracy, ConfusionMatrix, F1Score, MetricCollection, Precision, Recall


def get_classification_full_metrics(
    task: Literal["binary", "multiclass", "multilabel"] = "multiclass",
    threshold: float = 0.5,
    num_classes: Optional[int] = None,
    num_labels: Optional[int] = None,
    average_method: Literal["micro", "macro", "weighted", "none"] = "micro",
    multidim_average: Literal["global", "samplewise"] = "global",
    top_k: Optional[int] = 1,
    ignore_index: Optional[int] = None,
    validate_args: bool = True,
    prefix: Optional[str] = None,
    kwargs: dict = {},
    n_agent: int = 1,
) -> MetricCollection:
    return MetricCollection(
        [
            Accuracy(
                task=task,
                threshold=threshold,
                num_classes=num_classes,
                num_labels=num_labels,
                average=average_method,
                multidim_average=multidim_average,
                top_k=top_k,
                ignore_index=ignore_index,
                validate_args=validate_args,
                **kwargs,
            ),
            Precision(
                task=task,
                threshold=threshold,
                num_classes=num_classes,
                num_labels=num_labels,
                average=average_method,
                multidim_average=multidim_average,
                top_k=top_k,
                ignore_index=ignore_index,
                validate_args=validate_args,
                **kwargs,
            ),
            Recall(
                task=task,
                threshold=threshold,
                num_classes=num_classes,
                num_labels=num_labels,
                average=average_method,
                multidim_average=multidim_average,
                top_k=top_k,
                ignore_index=ignore_index,
                validate_args=validate_args,
                **kwargs,
            ),
            F1Score(
                task=task,
                threshold=threshold,
                num_classes=num_classes,
                num_labels=num_labels,
                average=average_method,
                multidim_average=multidim_average,
                top_k=top_k,
                ignore_index=ignore_index,
                validate_args=validate_args,
                **kwargs,
            ),
        ],
        prefix=prefix,
    )


def get_confusion_matrix(
    task: Literal["binary", "multiclass", "multilabel"] = "multiclass",
    threshold: float = 0.5,
    num_classes: Optional[int] = None,
    num_labels: Optional[int] = None,
    normalize: Optional[Literal["true", "pred", "all", "none"]] = None,
    ignore_index: Optional[int] = None,
    validate_args: bool = True,
    kwargs: dict = {},
) -> ConfusionMatrix:
    return ConfusionMatrix(
        task=task,
        threshold=threshold,
        num_classes=num_classes,
        num_labels=num_labels,
        normalize=normalize,
        ignore_index=ignore_index,
        validate_args=validate_args,
        **kwargs,
    )



from rlearn.sports.soccer.modules.seq2seq_encoder.pytoch_transformer_encoder import PytorchTransformerEncoder  # noqa: F401
from rlearn.sports.soccer.modules.seq2seq_encoder.pytorch_seq2seq_wrapper import (  # noqa: F401
    GruSeq2SeqEncoder,
    LstmSeq2SeqEncoder,
    RnnSeq2SeqEncoder,
)
from rlearn.sports.soccer.modules.seq2seq_encoder.seq2seq_encoder import Seq2SeqEncoder  # noqa: F401



# from /home/c_yeung/workspace6/python/openstarlab/Event/event/sports/soccer/main_class_soccer/main.py
from .soccer.main_class_soccer.main import rlearn_model_soccer

class RLearn_Model:
    def __new__(cls, *args, **kwargs):
        return rlearn_model_soccer(*args, **kwargs)


if __name__ == '__main__':
    import os
    # test split_data
    RLearn_Model(
        input_path=os.getcwd()+'/tests/data/datastadium/',
        output_path=os.getcwd()+'/tests/data/datastadium/split/'
    ).train_test_split()

    # test preprocess observation data
    RLearn_Model(
        config=os.getcwd()+'/tests/config/preprocessing_dssports2020.json',
        input_path=os.getcwd()+'/tests/data/datastadium/split/mini',
        output_path=os.getcwd()+'/tests/data/datastadium_simple_obs_action_seq/split/mini',
        num_process=5,
    ).preprocess_observation(batch_size=64)

    # test train model
    RLearn_Model(
        config=os.getcwd()+'/tests/config/exp_config.json'
    ).train(
        exp_name='sarsa_attacker',
        run_name='test',
        accelerator="gpu",
        devices=1,
        strategy="ddp",
    )

    # test visualize
    RLearn_Model().visualize_data(
        model_name='exp_config',
        checkpoint_path=os.getcwd()+'/rlearn/sports/output/sarsa_attacker/test/checkpoints/epoch=1-step=2.ckpt',
        match_id='2022100106',
        sequence_id=0,
    )

    print('Done')



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import matplotsoccer as mps
import os
from pathlib import Path
import japanize_matplotlib
from io import StringIO
import cv2
import ast
from sklearn.preprocessing import MinMaxScaler
import matplotlib.gridspec as gridspec
import glob
from ast import literal_eval
from math import pi
from matplotlib.colors import Normalize
import matplotlib.cm as cm

onball_action_names = [
    'pass',
    'through_pass',
    'shot',
    'cross',
    'dribble',
    'defense'
]

offball_action_names = [
    'idle',
    'up',
    'up_right',
    'right',
    'down_right',
    'down',
    'down_left',
    'left',
    'up_left'
]

offball_action_idx = [0, 1, 2, 3, 4, 5, 6, 7, 8]
onball_action_idx = [9, 10, 11, 12, 13, 14]


def preprocess_tracking_data(df, sequence_id):
    df_sequence = df[df['sequence_id'] == sequence_id].copy()
    df_sequence.loc[:, 'raw_state'] = df_sequence['events'].apply(lambda x: x['state'])

    df_sequence = df_sequence.drop(columns=['events'])
    df_sequence.reset_index(drop=True, inplace=True)

    df_sequence['ball_x'] = df_sequence['raw_state'].apply(lambda x: x['ball']['position']['x'])
    df_sequence['ball_y'] = df_sequence['raw_state'].apply(lambda x: x['ball']['position']['y'])


    df_sequence['attack_team_position'] = df_sequence['raw_state'].apply(lambda x: [player['position'] for player in x['attack_players']])
    df_sequence['attack_team_player'] = df_sequence['raw_state'].apply(lambda x: [player['player_name'] for player in x['attack_players']])
    df_sequence['attack_team_action'] = df_sequence['raw_state'].apply(lambda x: [player['action'] for player in x['attack_players']])
    df_sequence['defence_team_position'] = df_sequence['raw_state'].apply(lambda x: [player['position'] for player in x['defense_players']])
    df_sequence['defence_team_player'] = df_sequence['raw_state'].apply(lambda x: [player['player_name'] for player in x['defense_players']])

    return df_sequence


def safe_literal_eval(val):
    try:
        return ast.literal_eval(val)
    except (ValueError, SyntaxError):
        val = val.replace('nan', 'math.nan')
        val = eval(val)
        return val


def preprocess_q_values(q_values):
    q_values['q_value'] = q_values['q_value'].apply(lambda x: x[7:-1])
    q_values['q_value'] = q_values['q_value'].apply(lambda x: np.array(safe_literal_eval(x.replace('\n', ''))) if isinstance(x, str) else x)
    return q_values


def normalize_list(values):
    scaler = MinMaxScaler()
    values = np.array(values).reshape(-1, 1)
    return scaler.fit_transform(values).flatten()


def split_q_values(values, offball_action_idx, onball_action_idx): 
    offball_values = values[offball_action_idx] 
    onball_values = values[onball_action_idx] 
    return offball_values, onball_values


def offside(attackers, defenders):
    attackers_x = [attacker['x'] for attacker in attackers]
    offside_attackers = [0]*len(attackers)
    defenders_min_x = min([defender['x'] for defender in defenders])
    offside_line = 52.5 if defenders_min_x < 52.5 else defenders_min_x

    for i, attacker_x in enumerate(attackers_x):
        if attacker_x > offside_line:
            offside_attackers[i] = 1

    return offside_attackers


def plot_q_values(df_sequence, q_values, name, team_name, match_id, sequence_id):
    offball_directions = ["U", "U-L", "L", "B-L", "B", "B-R", "R", "U-R"]
    onball_actions = onball_action_names
    offball_q_values = 0
    angles_offball = np.linspace(0, 2 * np.pi, len(offball_directions) + 1)
    angles_onball = np.linspace(0, 2 * np.pi, len(onball_actions) + 1)

    cmap = cm.viridis  # colormap (e.g., viridis)
    yticks_offball, yticks_onball = [], []
    for idx, (data, q_value) in enumerate(zip(df_sequence.iterrows(), q_values.iterrows())):
        fig = plt.figure(figsize=(16, 9))
        gs = gridspec.GridSpec(2, 2, height_ratios=[1, 1], width_ratios=[1, 1])

        ax1 = fig.add_subplot(gs[:, 0])
        ax2 = fig.add_subplot(gs[:, 1], polar=True)
        mps.field(ax=ax1, show=False, color="green")
        ball_x, ball_y = data[1]['ball_x'] + 52.5, data[1]['ball_y'] + 34

        if ball_x - 25 < -2:
            field_x_min = -2
            field_x_max = 55
        elif ball_x + 25 > 107:
            field_x_min = 45
            field_x_max = 80
        else:
            field_x_min = ball_x-25
            field_x_max = ball_x+25
        ax1.plot(ball_x, ball_y, 'ko', markersize=7)
        ax1.set_xlim(field_x_min, field_x_max)
        ax1.set_ylim(-2, 70)
        ax1.set_title(f'{team_name}_{sequence_id}_{name}_{idx:04d}', fontsize=15)
        
        for position, player in zip(data[1]['attack_team_position'], data[1]['attack_team_player']):
            ax1.plot(position['x'] + 52.5, position['y'] + 34, 'ro', markersize=7)
            if player == name and field_x_min <= position['x'] + 52.5 <= field_x_max:
                ax1.text(position['x'] + 52, position['y'] + 35, player, fontsize=15)
        for position, player in zip(data[1]['defence_team_position'], data[1]['defence_team_player']):
            ax1.plot(position['x'] + 52.5, position['y'] + 34, 'bo', markersize=7)

        ax1.set_xticks([])
        ax1.set_yticks([])
        ax1.set_title(f'{data[1]["team_name_attack"]} vs {data[1]["team_name_defense"]}', fontsize=15)

        idle_q_values = q_value[1]['q_value_offball'][0].copy()
        offball_q_values = q_value[1]['q_value_offball'][1:].copy()

        offball_q_values = np.nan_to_num(offball_q_values, nan=0.0, posinf=0.0, neginf=0.0)
        idle_q_values = np.nan_to_num(idle_q_values, nan=0.0, posinf=0.0, neginf=0.0)

        ax2.set_xticks(angles_offball[:-1])
        ax2.set_xticklabels(offball_directions, color='grey', fontsize=15)
        yticks_offball = np.round(np.arange(-1, 1.2, 0.5), 2)
        ax2.set_yticks(yticks_offball)
        ax2.set_yticklabels([f'{ytick:.2f}' for ytick in yticks_offball], color='grey', fontsize=8)
        ax2.set_ylim(-1, max(offball_q_values)+0.3)        
        # ax2.set_theta_offset(np.pi / 2)
        ax2.set_title(f'Q Values for Offball Actions ({name})', fontsize=15)
        
        norm = Normalize(vmin=min(offball_q_values), vmax=max(offball_q_values))
        colors = cmap(norm(offball_q_values))

        for i in range(len(offball_q_values)):
            ax2.bar(
                angles_offball[i], 
                offball_q_values[i], 
                width=(angles_offball[i + 1]-0.2 - angles_offball[i]),
                color=colors[i],
                edgecolor='black',
                alpha=0.7,
            )

            ax2.text(
                angles_offball[i], 
                max(offball_q_values), 
                f'{offball_q_values[i]:.3g}'
            )

        scatter_size = max(idle_q_values * 1000, 0)
        ax2.scatter(0, -1, s=scatter_size, c='blue', alpha=0.7, label=f'{idle_q_values}')
        ax2.text(0, -1, f"{idle_q_values:.3g}", ha='center', va='center', color='white', fontsize=8)
        


        # onball_q_values = q_value[1]['q_value_onball'].copy()
        # norm = Normalize(vmin=min(onball_q_values), vmax=max(onball_q_values))
        # colors = cmap(norm(onball_q_values))

        # for i in range(len(onball_q_values)):
        #     ax3.bar(
        #         angles_onball[i],
        #         onball_q_values[i], 
        #         width=(angles_onball[i + 1]-0.2 - angles_onball[i]), 
        #         color=colors[i],
        #         edgecolor='black',
        #         alpha=0.7,
        #     )

        #     ax3.text(
        #         angles_onball[i],
        #         max(onball_q_values),
        #         f'{onball_q_values[i]:.3g}'
        #     )

        # ax3.set_xticks(angles_onball[:-1]) 
        # ax3.set_xticklabels(onball_actions, color='black', fontsize=10)
        # yticks_onball = np.round(np.arange(-1, 1.2, 0.5), 2)
        # # print(yticks_onball)
        # ax3.set_yticks(yticks_onball)
        # ax3.set_yticklabels([f'{ytick:.2f}' for ytick in yticks_onball], color='grey', fontsize=8)
        # ax3.set_ylim(-1, max(onball_q_values)+0.3)
        # ax3.set_title(f'Q Values for Onball Actions ({name})', fontsize=12)
        # ax3.set_theta_offset(np.pi/2)

        output_path = os.getcwd()+f'/tests/data/figures/{match_id}/'
        if not os.path.exists(output_path):
            os.makedirs(output_path)
        plt.tight_layout()
        plt.show()
        plt.savefig(output_path+f'frame_{name}_{sequence_id}_{idx:04d}.png')
        plt.close(fig)

        if idx == 10:
            break


def movie_from_images(image_files, output_file):
    frame = cv2.imread(image_files[0])
    height, width, _ = frame.shape
    video = cv2.VideoWriter(output_file, cv2.VideoWriter_fourcc(*'MJPG'), 1, (width, height))

    for image in image_files:
        video.write(cv2.imread(image))

    cv2.destroyAllWindows()
    video.release()

    for image in image_files:
        os.remove(image)


def create_movie(q_values_path, match_id, sequence_id):
    target_file_path = q_values_path
    tracking_data_path = os.getcwd()+f'/tests/data/datastadium/{match_id}/events.jsonl' # tracking data path preprocessed in SAR Package

    df = pd.DataFrame()
    data_list = []
    with open(tracking_data_path, 'r') as file:
        for line in file:
            data_list.append(pd.read_json(StringIO(line)))

    df = pd.concat(data_list, axis=0)

    df_sequence = preprocess_tracking_data(df, sequence_id)

    q_values = pd.read_csv(target_file_path)

    q_values = preprocess_q_values(q_values)

    # create new columns for offball and onball q_values
    q_values[['q_value_offball', 'q_value_onball']] = pd.DataFrame(q_values['q_value'].apply(lambda x: split_q_values(x, offball_action_idx, onball_action_idx)).tolist(), index=q_values.index)

    # standardize q_values
    q_values['q_value_offball'] = q_values['q_value_offball'].apply(normalize_list)
    q_values['q_value_onball'] = q_values['q_value_onball'].apply(normalize_list)

    player_name = q_values['player_name'].unique()
    print(player_name)
    for i in range(len(player_name)):
        name = player_name[i]
        team_name = q_values['team_name'][q_values['player_name'] == name].values[0]

        print(f"player name: {name}")
        plot_q_values(df_sequence, q_values, name, team_name, match_id, sequence_id)
        image_files = sorted(glob.glob(os.getcwd()+f'/tests/data/figures/{match_id}/frame_{name}_{sequence_id}*.png'))
        output_dir = os.getcwd()+f'/tests/data/movies/'
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        output_file = os.path.join(output_dir, f'{match_id}_{sequence_id}_{name}.avi')
        # movie_from_images(image_files, output_file)
        break



import logging
from copy import deepcopy
from itertools import chain
from pathlib import Path
from typing import Any, Dict, List, cast

import torch

from rlearn.sports.soccer.dataclass import (
    ObservationActionForLMBatch,
    ObservationactionForLMInstance,
    SimpleObservationAction,
    SimpleObservationActionSequence,
)
from rlearn.sports.soccer.source import (
    JLeagueSimpleObservationActionSequenceDataSource,
)
from rlearn.sports.soccer.env import DATA_DIR, PROJECT_DIR
from rlearn.sports.soccer.modules.datamodule.datamodule import DataModule
from rlearn.sports.soccer.modules.state_action_tokenizer.state_action_tokenizer import StateActionTokenizerBase
from rlearn.sports.soccer.utils.file_utils import load_jsonlines

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


@DataModule.register("jleague_observation_action_sequence")
class JLeagueSimpleObservationActionSequenceDataModule(DataModule):
    def __init__(
        self,
        state_action_tokenizer: Dict[str, Any],
        train_dataset: torch.utils.data.Dataset,
        valid_dataset: torch.utils.data.Dataset | None = None,
        test_dataset: torch.utils.data.Dataset | None = None,
        batch_size: int = 128,
        max_sequence_length: int = 4096,
        num_workers: int = 8,
    ) -> None:
        super().__init__()
        self.train_dataset = train_dataset
        self.valid_dataset = valid_dataset
        self.test_dataset = test_dataset
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.state_action_tokenizer = StateActionTokenizerBase.from_params(state_action_tokenizer)
        self.max_sequence_length = max_sequence_length

    @classmethod
    def _preprocess_data(cls, examples, state_action_tokenizer: StateActionTokenizerBase):
        sequence_in_examples = []
        action_mask_in_examples = []
        for sequence in examples['sequence']:
            sequence_in_example = []
            action_mask_in_example = []
            for observation_action in sequence:
                observation_action = SimpleObservationAction.from_dict(observation_action)
                # [SEP] position velocity position velocity ... [ACTION_SEP] action [SEP] position velocity position velocity ...
                sub_sequence = []
                sub_sequence.append(state_action_tokenizer.encode("[SEP]"))
                for player in observation_action.observation.players:
                    if state_action_tokenizer.encode_position_separately:
                        sub_sequence.extend(
                            [
                                state_action_tokenizer.encode(player.position.x),
                                state_action_tokenizer.encode(player.position.y),
                                state_action_tokenizer.encode(player.velocity),
                            ]
                        )
                    else:
                        sub_sequence.extend(
                            [
                                state_action_tokenizer.encode(player.position),
                                state_action_tokenizer.encode(player.velocity),
                            ]
                        )

                if state_action_tokenizer.encode_position_separately:
                    sub_sequence.extend(
                        [
                            state_action_tokenizer.encode(observation_action.observation.ball.position.x),
                            state_action_tokenizer.encode(observation_action.observation.ball.position.y),
                            state_action_tokenizer.encode(observation_action.observation.ball.velocity),
                            state_action_tokenizer.encode(observation_action.observation.ego_player.position.x),
                            state_action_tokenizer.encode(observation_action.observation.ego_player.position.y),
                        ]
                    )
                else:
                    sub_sequence.extend(
                        [
                            state_action_tokenizer.encode(observation_action.observation.ball.position),
                            state_action_tokenizer.encode(observation_action.observation.ball.velocity),
                            state_action_tokenizer.encode(observation_action.observation.ego_player.position),
                        ]
                    )
                sub_sequence.append(state_action_tokenizer.encode("[ACTION_SEP]"))
                sub_sequence.append(state_action_tokenizer.encode(observation_action.action))
                sequence_in_example.extend(sub_sequence)
                action_mask_in_example.extend([0] * (len(sub_sequence) - 1) + [1])
            sequence_in_examples.append(sequence_in_example)
            action_mask_in_examples.append(action_mask_in_example)
        assert len(sequence) != 0, f"sequence is empty: {sequence}"
        return {
            "input_ids": sequence_in_examples,
            "action_mask": action_mask_in_examples,
        }

    @classmethod
    def preprocess_data(
        cls,
        dataset: torch.utils.data.Dataset,
        state_action_tokenizer: Dict[str, Any],
        num_workers: int = 8,
        preprocess_batch_size: int = 32,
    ) -> torch.utils.data.Dataset:
        state_action_tokenizer = StateActionTokenizerBase.from_params(deepcopy(state_action_tokenizer))
        dataset = dataset.map(
            cls._preprocess_data,
            batched=True,
            num_proc=num_workers,
            batch_size=preprocess_batch_size,
            fn_kwargs={
                "state_action_tokenizer": state_action_tokenizer,
            },
        )
        # dataset = dataset.map(
        #     self._truncate_data,
        #     batched=True,
        #     num_proc=self.num_workers,
        # )
        return dataset

    # def _truncate_data(self, examples):
    #     concatenated_examples = {k: list(chain(*examples[k])) for k in examples.keys()}
    #     total_length = len(concatenated_examples[list(examples.keys())[0]])
    #     # We drop the small remainder, and if the total_length < max_seq_len  we exclude this batch and return an empty dict. # noqa
    #     # We could add padding if the model supported it instead of this drop, you can customize this part to your needs. # noqa
    #     total_length = (total_length // self.max_sequence_length) * self.max_sequence_length
    #     # Split by chunks of max_len.
    #     result = {
    #         k: [t[i : i + self.max_sequence_length] for i in range(0, total_length, self.max_sequence_length)]
    #         for k, t in concatenated_examples.items()
    #     }
    #     return result

    def train_dataloader(
        self,
        batch_size: int | None = None,
        shuffle: bool = True,
        num_workers: int | None = None,
    ) -> torch.utils.data.DataLoader:
        return self.build_dataloader(
            dataset=self.train_dataset,
            batch_size=batch_size or self.batch_size,
            shuffle=shuffle,
            num_workers=num_workers or self.num_workers,
        )

    def val_dataloader(
        self,
        batch_size: int | None = None,
        shuffle: bool = False,
        num_workers: int | None = None,
    ) -> torch.utils.data.DataLoader:
        assert self.valid_dataset is not None, f"valid dataset is not found: {self.valid_dataset}"
        return self.build_dataloader(
            dataset=self.valid_dataset,
            batch_size=batch_size or self.batch_size,
            shuffle=shuffle,
            num_workers=num_workers or self.num_workers,
        )

    def test_dataloader(
        self,
        batch_size: int | None = None,
        shuffle: bool = False,
        num_workers: int | None = None,
    ) -> torch.utils.data.DataLoader:
        assert self.test_dataset is not None, f"test dataset is not found: {self.test_dataset}"
        return self.build_dataloader(
            dataset=self.test_dataset,
            batch_size=batch_size or self.batch_size,
            shuffle=shuffle,
            num_workers=num_workers or self.num_workers,
        )

    def build_dataloader(
        self,
        dataset: torch.utils.data.Dataset,
        batch_size: int | None = None,
        shuffle: bool = False,
        num_workers: int = 0,
    ) -> torch.utils.data.DataLoader:
        return torch.utils.data.DataLoader(
            dataset=dataset,
            batch_size=batch_size or self.batch_size,
            shuffle=shuffle,
            collate_fn=self.batch_collator,
            num_workers=self.num_workers,
        )

    def batch_collator(self, instances) -> Dict[str, torch.Tensor]:
        max_length = max(len(instance['input_ids']) for instance in instances)
        input_ids = cast(
            torch.LongTensor, torch.full((len(instances), max_length), self.state_action_tokenizer.encode("[PAD]"))
        )
        mask = torch.zeros((len(instances), max_length), dtype=torch.long)
        action_mask = torch.zeros((len(instances), max_length), dtype=torch.long)

        for i, instance in enumerate(instances):
            length = len(instance["input_ids"])
            input_ids[i, :length] = torch.tensor(instance["input_ids"])
            mask[i, :length] = 1
            action_mask[i, :length] = torch.tensor(instance['action_mask'])

        batch = {
            "input_ids": input_ids,
            "mask": mask,
            "action_mask": action_mask,
        }
        return batch


# @DataModule.register("jleague_observation_action_sequence")
# class JLeagueSimpleObservationActionSequenceDataModule(DataModule):
#     def __init__(
#         self,
#         state_action_tokenizer: Dict[str, Any],
#         train_dataset,
#         valid_dataset,
#         test_dataset,
#         batch_size: int = 128,
#         max_sequence_length: int = 4096,
#         num_workers: int = 8,
#     ) -> None:
#         super().__init__()
#         self.train_dataset = train_dataset
#         self.valid_dataset = valid_dataset
#         self.test_dataset = test_dataset
#         self.batch_size = batch_size
#         self.num_workers = num_workers
#         self.state_action_tokenizer = StateActionTokenizerBase.from_params(state_action_tokenizer)
#         self.max_sequence_length = max_sequence_length

#     def _prepare_data(self, dataset):
#         dataset = dataset.map(
#             self._preprocess_data,
#             batched=True,
#             num_proc=self.num_workers,
#             remove_columns=dataset.column_names,
#         )
#         dataset = dataset.map(
#             self._truncate_data,
#             batched=True,
#             num_proc=self.num_workers,
#         )
#         return dataset

#     def _truncate_data(self, examples):
#         # Concatenate all texts.
#         concatenated_examples = {k: list(chain(*examples[k])) for k in examples.keys()} # noqa
#         total_length = len(concatenated_examples[list(examples.keys())[0]])
#         # We drop the small remainder, and if the total_length < max_seq_len  we exclude this batch and return an empty dict. # noqa
#         # We could add padding if the model supported it instead of this drop, you can customize this part to your needs. # noqa
#         total_length = (total_length // self.max_sequence_length) * self.max_sequence_length
#         # Split by chunks of max_len.
#         result = {
#             k: [t[i : i + self.max_sequence_length] for i in range(0, total_length, self.max_sequence_length)]
#             for k, t in concatenated_examples.items()
#         }
#         return result

#     def _preprocess_data(self, examples):
#         sequence_in_examples = []
#         action_mask_in_examples = []
#         for sequence in examples['sequence']:
#             sequence_in_example = []
#             action_mask_in_example = []
#             for observation_action in sequence:
#                 observation_action = SimpleObservationAction.from_dict(observation_action)
#                 # [SEP] position velocity position velocity ... [ACTION_SEP] action [SEP] position velocity position velocity ...
#                 sub_sequence = []
#                 sub_sequence.append(self.state_action_tokenizer.encode("[SEP]"))
#                 for player in observation_action.observation.players:
#                     if self.state_action_tokenizer.encode_position_separately:
#                         sub_sequence.extend(
#                             [
#                                 self.state_action_tokenizer.encode(player.position.x),
#                                 self.state_action_tokenizer.encode(player.position.y),
#                                 self.state_action_tokenizer.encode(player.velocity),
#                             ]
#                         )
#                     else:
#                         sub_sequence.extend(
#                             [
#                                 self.state_action_tokenizer.encode(player.position),
#                                 self.state_action_tokenizer.encode(player.velocity),
#                             ]
#                         )

#                 if self.state_action_tokenizer.encode_position_separately:
#                     sub_sequence.extend(
#                         [
#                             self.state_action_tokenizer.encode(observation_action.observation.ball.position.x),
#                             self.state_action_tokenizer.encode(observation_action.observation.ball.position.y),
#                             self.state_action_tokenizer.encode(observation_action.observation.ball.velocity),
#                         ]
#                     )
#                 else:
#                     sub_sequence.extend(
#                         [
#                             self.state_action_tokenizer.encode(observation_action.observation.ball.position),
#                             self.state_action_tokenizer.encode(observation_action.observation.ball.velocity),
#                         ]
#                     )
#                 sub_sequence.append(self.state_action_tokenizer.encode("[ACTION_SEP]"))
#                 sub_sequence.append(self.state_action_tokenizer.encode(observation_action.action))
#                 sequence_in_example.extend(sub_sequence)
#                 action_mask_in_example.extend([0] * (len(sub_sequence) - 1) + [1])
#             sequence_in_examples.append(sequence_in_example)
#             action_mask_in_examples.append(action_mask_in_example)
#         assert len(sequence) != 0, f"sequence is empty: {sequence}"
#         return {
#             "input_ids": sequence_in_examples,
#             "action_mask": action_mask_in_examples,
#         }

#     def train_dataloader(
#         self,
#         batch_size: int | None = None,
#         shuffle: bool = True,
#         num_workers: int | None = None,
#     ) -> torch.utils.data.DataLoader:
#         return self.build_dataloader(
#             dataset=self.train_dataset,
#             batch_size=batch_size or self.batch_size,
#             shuffle=shuffle,
#             num_workers=num_workers or self.num_workers,
#         )

#     def val_dataloader(
#         self,
#         batch_size: int | None = None,
#         shuffle: bool = False,
#         num_workers: int | None = None,
#     ) -> torch.utils.data.DataLoader:
#         assert self.valid_dataset is not None, f"valid dataset is not found: {self.dataset}"
#         return self.build_dataloader(
#             dataset=self.valid_dataset,
#             batch_size=batch_size or self.batch_size,
#             shuffle=shuffle,
#             num_workers=num_workers or self.num_workers,
#         )

#     def test_dataloader(
#         self,
#         batch_size: int | None = None,
#         shuffle: bool = False,
#         num_workers: int | None = None,
#     ) -> torch.utils.data.DataLoader:
#         assert self.test_dataset is not None, f"test dataset is not found: {self.dataset}"
#         return self.build_dataloader(
#             dataset=self.test_dataset,
#             batch_size=batch_size or self.batch_size,
#             shuffle=shuffle,
#             num_workers=num_workers or self.num_workers,
#         )

#     def build_dataloader(
#         self,
#         dataset: torch.utils.data.Dataset,
#         batch_size: int | None = None,
#         shuffle: bool = False,
#         num_workers: int = 0,
#     ) -> torch.utils.data.DataLoader:
#         return torch.utils.data.DataLoader(
#             dataset=dataset,
#             batch_size=batch_size or self.batch_size,
#             shuffle=shuffle,
#             collate_fn=self.batch_collator,
#             num_workers=self.num_workers,
#         )

#     def batch_collator(self, instances) -> Dict[str, torch.Tensor]:
#         max_length = max(len(instance['input_ids']) for instance in instances)
#         input_ids = cast(
#             torch.LongTensor, torch.full((len(instances), max_length), self.state_action_tokenizer.encode("[PAD]"))
#         )
#         mask = torch.zeros((len(instances), max_length), dtype=torch.long)
#         action_mask = torch.zeros((len(instances), max_length), dtype=torch.long)

#         for i, instance in enumerate(instances):
#             length = len(instance["input_ids"])
#             input_ids[i, :length] = torch.tensor(instance["input_ids"])
#             mask[i, :length] = 1
#             action_mask[i, :length] = torch.tensor(instance['action_mask'])

#         return {
#             "input_ids": input_ids,
#             "mask": mask,
#             "action_mask": action_mask,
#         }
